//
// File-system system calls.
// Mostly argument checking, since we don't trust
// user code, and calls into file.c and fs.c.
//


#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/stat.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/sleeplock.h"
#include "include/file.h"
#include "include/pipe.h"
#include "include/fcntl.h"
#include "include/fat32.h"
#include "include/syscall.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/vm.h"
#include "include/kalloc.h"
#include "include/memlayout.h"

#define PROT_NONE 0
#define PROT_READ 1
#define PROT_WRITE 2
#define PROT_EXEC 4

#define MAP_FILE 0
#define MAP_SHARED 0x01
#define MAP_PRIVATE 0X02
#define MAP_FAILED ((void *) -1)

// Fetch the nth word-sized system call argument as a file descriptor
// and return both the descriptor and the corresponding struct file.
static int
argfd(int n, int *pfd, struct file **pf)
{
  int fd;
  struct file *f;

  if(argint(n, &fd) < 0)
    return -1;
  if(fd < 0 || fd >= NOFILE || (f=myproc()->ofile[fd]) == NULL)
    return -1;
  if(pfd)
    *pfd = fd;
  if(pf)
    *pf = f;
  return 0;
}

// Allocate a file descriptor for the given file.
// Takes over file reference from caller on success.
static int
fdalloc(struct file *f)
{
  int fd;
  struct proc *p = myproc();

  for(fd = 0; fd < NOFILE; fd++){
    if(p->ofile[fd] == 0){
      p->ofile[fd] = f;
      return fd;
    }
  }
  return -1;
}

uint64
sys_dup(void)
{
  struct file *f;
  int fd;

  if(argfd(0, 0, &f) < 0)
    return -1;
  if((fd=fdalloc(f)) < 0)
    return -1;
  filedup(f);
  return fd;
}

uint64
sys_dup2(void)
{
  struct file *oldfd;
  int newfd;
  // printf("I'm testing dup2\n");
  // if(argfd(0, 0, &oldfd) < 0||argint(1,&newfd) < 0||newfd>=NOFILE)
  //   return -1;
  if(argfd(0, 0, &oldfd) < 0)
    return -1;
  // printf("get oldfd\n");
  if(argint(1,&newfd) < 0)
    return -1;
  // printf("get newfd\n");
  if(newfd>=NOFILE)
    return -1;
  // printf("I have got proper parameters\n");
  struct proc *p=myproc();
  if(p->ofile[newfd]!=0) {
    // printf("start close file\n");
    fileclose(p->ofile[newfd]);
    myproc()->ofile[newfd]=0;
  }
  p->ofile[newfd] = oldfd;
  // printf("start dup file\n");
  filedup(oldfd);
  return newfd;
}

uint64
sys_read(void)
{
  struct file *f;
  int n;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
    return -1;
  return fileread(f, p, n);
}

uint64
sys_write(void)
{
  struct file *f;
  int n;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
    return -1;

  return filewrite(f, p, n);
}

uint64
sys_close(void)
{
  int fd;
  struct file *f;

  if(argfd(0, &fd, &f) < 0)
    return -1;
  myproc()->ofile[fd] = 0;
  fileclose(f);
  return 0;
}

uint64
sys_fstat(void)
{
  struct file *f;
  uint64 st; // user pointer to struct stat

  if(argfd(0, 0, &f) < 0 || argaddr(1, &st) < 0)
    return -1;
  return filestat(f, st);
}

uint64
sys_fstat2(void)
{
  struct file *f;
  uint64 st; // user pointer to struct stat

  if(argfd(0, 0, &f) < 0 || argaddr(1, &st) < 0)  return -1;
  filestat(f, st);
  struct stat fstat2_res;
  copyin2((char *)&fstat2_res, st, sizeof(fstat2_res));
  struct kstat f_kstat;
  f_kstat.st_dev = fstat2_res.dev;
  f_kstat.st_ino = 18;
  f_kstat.st_mode = fstat2_res.type;
  f_kstat.st_nlink = f->ref;
  f_kstat.st_uid = 0;
  f_kstat.st_gid = 1;
  f_kstat.st_rdev = 0;
  f_kstat.__pad = 0;
  f_kstat.st_size = fstat2_res.size;
  f_kstat.st_blksize = 0;
  f_kstat.__pad2 = 0;
  f_kstat.st_blocks = 0;
  f_kstat.st_atime_sec = 0;
  f_kstat.st_atime_nsec = 0;
  f_kstat.st_mtime_sec = 0;
  f_kstat.st_mtime_nsec = 0;
  f_kstat.st_ctime_sec = 0;
  f_kstat.st_ctime_nsec = 0;
  f_kstat.__unused[0] = 0;
  f_kstat.__unused[1] = 0;
  return copyout2(st, (char *)&f_kstat, sizeof(f_kstat)) < 0;
}

static struct dirent*
create(char *path, short type, int mode)
{
  struct dirent *ep, *dp;
  char name[FAT32_MAX_FILENAME + 1];

  if((dp = enameparent(path, name)) == NULL)
    return NULL;

  if (type == T_DIR) {
    mode = ATTR_DIRECTORY;
  } else if (mode & O_RDONLY) {
    mode = ATTR_READ_ONLY;
  } else {
    mode = 0;  
  }

  elock(dp);
  if ((ep = ealloc(dp, name, mode)) == NULL) {
    eunlock(dp);
    eput(dp);
    return NULL;
  }
  
  if ((type == T_DIR && !(ep->attribute & ATTR_DIRECTORY)) ||
      (type == T_FILE && (ep->attribute & ATTR_DIRECTORY))) {
    eunlock(dp);
    eput(ep);
    eput(dp);
    return NULL;
  }

  eunlock(dp);
  eput(dp);

  elock(ep);
  return ep;
}

uint64 myOpen(char path[FAT32_MAX_PATH], int omode)
{
    int fd;
    struct file *f;
    struct dirent *ep;
    if ((omode & O_CREATE) || (omode & 0x40))
    {
        ep = create(path, T_FILE, omode);
        if (ep == NULL)
        {
            return -1;
        }
    }
    else
    {
        if ((ep = ename(path)) == NULL)
        {
            return -1;
        }
        elock(ep);
        if (ep->attribute & ATTR_DIRECTORY)
        {
            omode = O_RDONLY;
        }
        if ((ep->attribute & ATTR_DIRECTORY) && omode != O_RDONLY)
        {
            eunlock(ep);
            eput(ep);
            return -1;
        }
    }

    if ((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0)
    {
        if (f)
        {
            fileclose(f);
        }
        eunlock(ep);
        eput(ep);
        return -1;
    }

    if (!(ep->attribute & ATTR_DIRECTORY) && (omode & O_TRUNC))
    {
        etrunc(ep);
    }

    f->type = FD_ENTRY;
    f->off = (omode & O_APPEND) ? ep->file_size : 0;
    f->ep = ep;
    f->readable = !(omode & O_WRONLY);
    f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

    eunlock(ep);

    return fd;
}

uint64
sys_open(void)
{
  char path[FAT32_MAX_PATH];
  int omode;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argint(1, &omode) < 0)
    return -1;

  return myOpen(path,omode);
}

// 输出信息
// init: starting sh
// executing test mkdir
// mkdir没有进入测试
uint64
sys_mkdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;

  printf("I'm testing mkdir");

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = create(path, T_DIR, 0)) == 0){
    return -1;
  }
  eunlock(ep);
  eput(ep);
  return 0;
}

uint64
sys_mkdir_(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;

  printf("I'm testing mkdir_\n");

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = create(path, T_DIR, 0)) == 0){
    return -1;
  }
  eunlock(ep);
  eput(ep);
  return 0;
}

uint64
sys_mkdirat(void)
{
    char path[FAT32_MAX_PATH];
  struct dirent *ep;
  int omode;
  int dirfd;
  
  if(argint(0,&dirfd)<0||argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &omode) < 0)
    return -1;

  if(dirfd!=-100){
    struct proc *p = myproc();
    int n=strlen(p->ofile[dirfd]->ep->filename);
    int m=strlen(path);
    for(int i=m;i>=0;--i)path[i+n]=path[i];
    for(int i=0;i<n;++i)path[i]=p->ofile[dirfd]->ep->filename[i];
  }  
  if((ep = create(path, T_DIR, omode)) == 0)return -1;
  eunlock(ep);
  eput(ep);
  return 0;
  // char path[FAT32_MAX_PATH];
  // struct dirent *ep;
  // int omode;
  // int dirfd;
  
  // printf("I'm testing mkdirat\n");

  // if(argint(0,&dirfd)<0||argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &omode) < 0)
  //   return -1;
  // if((ep = create(path, T_DIR, 0)) == 0)
  //   return -1;
  // eunlock(ep);
  // eput(ep);
  // return 0;
  // 不考虑model
  // if(dirfd!=-100){
  //   struct proc *p = myproc();
  //   int n=strlen(p->ofile[dirfd]->ep->filename);
  //   int m=strlen(path);
  //   for(int i=m;i>=0;--i)
  //     path[i+n]=path[i];
  //   for(int i=0;i<n;++i)
  //     path[i]=p->ofile[dirfd]->ep->filename[i];
  // }  
  // if((ep = create(path, T_DIR, omode)) == 0)
  //   return -1;
  // eunlock(ep);
  // eput(ep);
  // return 0;
}

uint64
sys_chdir(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  struct proc *p = myproc();
  
  if(argstr(0, path, FAT32_MAX_PATH) < 0 || (ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if(!(ep->attribute & ATTR_DIRECTORY)){
    eunlock(ep);
    eput(ep);
    return -1;
  }
  eunlock(ep);
  eput(p->cwd);
  p->cwd = ep;
  return 0;
}

uint64
sys_pipe(void)
{
  uint64 fdarray; // user pointer to array of two integers
  struct file *rf, *wf;
  int fd0, fd1;
  struct proc *p = myproc();

  if(argaddr(0, &fdarray) < 0)
    return -1;
  if(pipealloc(&rf, &wf) < 0)
    return -1;
  fd0 = -1;
  if((fd0 = fdalloc(rf)) < 0 || (fd1 = fdalloc(wf)) < 0){
    if(fd0 >= 0)
      p->ofile[fd0] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  if(copyout2(fdarray, (char*)&fd0, sizeof(fd0)) < 0 ||
     copyout2(fdarray+sizeof(fd0), (char *)&fd1, sizeof(fd1)) < 0){
    p->ofile[fd0] = 0;
    p->ofile[fd1] = 0;
    fileclose(rf);
    fileclose(wf);
    return -1;
  }
  return 0;
}

// To open console device.
uint64
sys_dev(void)
{
  int fd, omode;
  int major, minor;
  struct file *f;

  if(argint(0, &omode) < 0 || argint(1, &major) < 0 || argint(2, &minor) < 0){
    return -1;
  }

  if(omode & O_CREATE){
    panic("dev file on FAT");
  }

  if(major < 0 || major >= NDEV)
    return -1;

  if((f = filealloc()) == NULL || (fd = fdalloc(f)) < 0){
    if(f)
      fileclose(f);
    return -1;
  }

  f->type = FD_DEVICE;
  f->off = 0;
  f->ep = 0;
  f->major = major;
  f->readable = !(omode & O_WRONLY);
  f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

  return fd;
}

// To support ls command
uint64
sys_readdir(void)
{
  struct file *f;
  uint64 p;

  if(argfd(0, 0, &f) < 0 || argaddr(1, &p) < 0)
    return -1;
  return dirnext(f, p);
}

// absolute cwd
uint64
sys_getcwd(void)
{
    uint64 addr;
    if (argaddr(0, &addr) < 0)
        return -1;

    struct dirent *de = myproc()->cwd;
    char path[FAT32_MAX_PATH];
    char *s;
    int len;

    if (de->parent == NULL)
    {
        s = "/";
    }
    else
    {
        s = path + FAT32_MAX_PATH - 1;
        *s = '\0';
        while (de->parent)
        {
            len = strlen(de->filename);
            s -= len;
            if (s <= path) // can't reach root "/"
                return -1;
            strncpy(s, de->filename, len);
            *--s = '/';
            de = de->parent;
        }
    }
    if (copyout2(addr, s, strlen(s) + 1) < 0)
        return -1;

    return addr;
}

// uint64
// sys_getcwd(void)
// {
//   uint64 addr;
//   if (argaddr(0, &addr) < 0)
//     return -1;

//   struct dirent *de = myproc()->cwd;
//   char path[FAT32_MAX_PATH];
//   char *s;
//   int len;
//   int path_len = 0;
//   if (de == NULL) { //de:myproc()->cwd
//     return -1; // 当前目录项为空
//   }
//   s = path + FAT32_MAX_PATH - 1;
//   *s = '\0'; // 设置路径字符串的结束符
//   // 逆向构建路径
//   while (de) {
//     len = strlen(de->filename);
//     path_len += len + 1; // 计算路径长度
//     if (path_len > FAT32_MAX_PATH) // 检查路径是否超出最大长度
//       return -1;
//     s -= len;
//     strncpy(s, de->filename, len);
//     *--s = '/'; // 添加 '/' 作为分隔符
//     de = de->parent;
//   }
//   // 删除开头多余的 '/'
//   if (s[0] == '/') {
//     s++; // 移动到下一个字符
//   }

//   // 拷贝路径到用户空间
//   if (copyout2(addr, s, path + FAT32_MAX_PATH - s) < 0)
//     return -1;

//   return 0;
// }


// Is the directory dp empty except for "." and ".." ?
static int
isdirempty(struct dirent *dp)
{
  struct dirent ep;
  int count;
  int ret;
  ep.valid = 0;
  ret = enext(dp, &ep, 2 * 32, &count);   // skip the "." and ".."
  return ret == -1;
}

uint64
sys_remove(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  int len;
  if((len = argstr(0, path, FAT32_MAX_PATH)) <= 0)
    return -1;

  char *s = path + len - 1;
  while (s >= path && *s == '/') {
    s--;
  }
  if (s >= path && *s == '.' && (s == path || *--s == '/')) {
    return -1;
  }
  
  if((ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if((ep->attribute & ATTR_DIRECTORY) && !isdirempty(ep)){
      eunlock(ep);
      eput(ep);
      return -1;
  }
  elock(ep->parent);      // Will this lead to deadlock?
  eremove(ep);
  eunlock(ep->parent);
  eunlock(ep);
  eput(ep);

  return 0;
}

// Must hold too many locks at a time! It's possible to raise a deadlock.
// Because this op takes some steps, we can't promise
uint64
sys_rename(void)
{
  char old[FAT32_MAX_PATH], new[FAT32_MAX_PATH];
  if (argstr(0, old, FAT32_MAX_PATH) < 0 || argstr(1, new, FAT32_MAX_PATH) < 0) {
      return -1;
  }

  struct dirent *src = NULL, *dst = NULL, *pdst = NULL;
  int srclock = 0;
  char *name;
  if ((src = ename(old)) == NULL || (pdst = enameparent(new, old)) == NULL
      || (name = formatname(old)) == NULL) {
    goto fail;          // src doesn't exist || dst parent doesn't exist || illegal new name
  }
  for (struct dirent *ep = pdst; ep != NULL; ep = ep->parent) {
    if (ep == src) {    // In what universe can we move a directory into its child?
      goto fail;
    }
  }

  uint off;
  elock(src);     // must hold child's lock before acquiring parent's, because we do so in other similar cases
  srclock = 1;
  elock(pdst);
  dst = dirlookup(pdst, name, &off);
  if (dst != NULL) {
    eunlock(pdst);
    if (src == dst) {
      goto fail;
    } else if (src->attribute & dst->attribute & ATTR_DIRECTORY) {
      elock(dst);
      if (!isdirempty(dst)) {    // it's ok to overwrite an empty dir
        eunlock(dst);
        goto fail;
      }
      elock(pdst);
    } else {                    // src is not a dir || dst exists and is not an dir
      goto fail;
    }
  }

  if (dst) {
    eremove(dst);
    eunlock(dst);
  }
  memmove(src->filename, name, FAT32_MAX_FILENAME);
  emake(pdst, src, off);
  if (src->parent != pdst) {
    eunlock(pdst);
    elock(src->parent);
  }
  eremove(src);
  eunlock(src->parent);
  struct dirent *psrc = src->parent;  // src must not be root, or it won't pass the for-loop test
  src->parent = edup(pdst);
  src->off = off;
  src->valid = 1;
  eunlock(src);

  eput(psrc);
  if (dst) {
    eput(dst);
  }
  eput(pdst);
  eput(src);

  return 0;

fail:
  if (srclock)
    eunlock(src);
  if (dst)
    eput(dst);
  if (pdst)
    eput(pdst);
  if (src)
    eput(src);
  return -1;
}

uint64
sys_openat(void)
{
  char path[FAT32_MAX_PATH];
  int omode;

  if(argstr(1, path, FAT32_MAX_PATH) < 0 || argint(2, &omode) < 0)
    return -1;
  
  return myOpen(path,omode);
}

uint64 
find_free_vm_for_mmap(uint64 page_need){
  uint64 ans;
  struct proc* p= myproc();
  uint64 used_space[NOFILE][2] = {};
  for(int j = 0; j < NOFILE; j++){
    used_space[j][0] = p->VMA_list[j].start_in_vm;
    used_space[j][1] = p->VMA_list[j].used_page_num;
  }
  // 选择排序
  for(int j = 0; j < NOFILE; j++){
    for(int k = j+1; k < NOFILE; k++){
      if(used_space[j][0] < used_space[k][0]){
        uint64 temp = used_space[j][0];
        used_space[j][0] = used_space[k][0];
        used_space[k][0] = temp;
        temp = used_space[j][1];
        used_space[j][1] = used_space[k][1];
        used_space[k][1] = temp;
      }
    }
  }

  if(used_space[0][0] == 0){
    ans = TRAPFRAME - PGSIZE - PGSIZE;
  } else {
    uint64 start_in_vm = 0;
    for(int j = 1; j < NOFILE; j++){
      uint64 pre_bottom = used_space[j-1][0];
      uint64 now_top = used_space[j][0] + PGSIZE * used_space[j][1];
      int now_gap = (pre_bottom - now_top) / PGSIZE;
      if(now_gap >= page_need){
        start_in_vm = now_top;
        break;      
      }
    }
    ans = start_in_vm;
    if (start_in_vm <= p->sz){  // 没有足够空间 
      ans =  -1;
    }
  }
  return ans;
}

uint64
sys_mmap(void) {
  uint64 start;
  int len;
  uint off;
  int port, flags, fd;
  struct file *F;
  if(argaddr(0, &start) < 0||argint(1, &len) < 0||argint(2, &port) < 0||argint(3, &flags) < 0||argfd(4, &fd,&F) < 0||argint(5, (int*)&off) < 0) {
    return -1;
  }
  struct proc *p=myproc();
  if(start==NULL)
  {
    int newSz=uvmalloc(p->pagetable,p->kpagetable,p->sz,p->sz+len);
    start=p->sz;
    p->sz=newSz;
  }
  F->off=0;//文件指针放到文件首部
  fileread(F,start,len);
  return start;
}

uint64
sys_munmap(void){
  return 0;
}

uint64
sys_getdents(void)
{
  struct file *f;
  int fd;
  uint64 addr;
  int size;
  if(argfd(0,&fd,&f)<0||argaddr(1,&addr)<0||argint(2,&size)<0)
    return -1;
  return getdents(f->ep,addr,size);
}

uint64
sys_unlink(void)
{
  char path[FAT32_MAX_PATH];
  struct dirent *ep;
  int len;
  int dirfd;
  if(argint(0,&dirfd)<0)return -1;
  if((len = argstr(1, path, FAT32_MAX_PATH)) <= 0)
    return -1;
  if(dirfd!=-100){
    struct proc *p = myproc();
    int n=strlen(p->ofile[dirfd]->ep->filename);
    int m=strlen(path);
    for(int i=m;i>=0;--i)path[i+n]=path[i];
    for(int i=0;i<n;++i)path[i]=p->ofile[dirfd]->ep->filename[i];
  }
  else {}
  char *s = path + len - 1;
  while (s >= path && *s == '/') {
    s--;
  }
  if (s >= path && *s == '.' && (s == path || *--s == '/')) {
    return -1;
  }
  if((ep = ename(path)) == NULL){
    return -1;
  }
  elock(ep);
  if((ep->attribute & ATTR_DIRECTORY) && !isdirempty(ep)){
      eunlock(ep);
      eput(ep);
      return -1;
  }
  elock(ep->parent);      // Will this lead to deadlock?
  eremove(ep);
  eunlock(ep->parent);
  eunlock(ep);
  eput(ep);

  return 0;
}

uint64
sys_mount(void){
  return 0;
}

uint64 
sys_umount(void){
  return 0;
}
