
#include "include/types.h"
#include "include/riscv.h"
#include "include/param.h"
#include "include/memlayout.h"
#include "include/spinlock.h"
#include "include/proc.h"
#include "include/syscall.h"
#include "include/timer.h"
#include "include/time.h"
#include "include/kalloc.h"
#include "include/string.h"
#include "include/printf.h"
#include "include/vm.h"
// #include "proc.c"

extern struct proc *waitpid_queue;
extern struct proc proc[NPROC];
extern struct spinlock pid_lock;
extern int exec(char *path, char **argv);

uint64
sys_exec(void)
{
  char path[FAT32_MAX_PATH], *argv[MAXARG];
  int i;
  uint64 uargv, uarg;

  if(argstr(0, path, FAT32_MAX_PATH) < 0 || argaddr(1, &uargv) < 0){
    return -1;
  }
  memset(argv, 0, sizeof(argv));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0)
      goto bad;
    if(fetchstr(uarg, argv[i], PGSIZE) < 0)
      goto bad;
  }

  int ret = exec(path, argv);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);

  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  return -1;
}

uint64 sys_execve(void){
  return sys_exec();
}

uint64
sys_exit(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}

uint64
sys_getppid(void)
{
  return myproc()->parent ? myproc()->parent->pid : 0;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64 
sys_clone(void){
  uint64 stack;//共享栈地址
  int flages;//标志
  if(argint(0,&flages)<0||argaddr(1,&stack)<0)
    return -1;
  if(stack==0)
    return fork();//不共享栈
  else
    return clone(stack);
}

uint64
sys_wait(void)
{
  uint64 p;
  if(argaddr(0, &p) < 0)
    return -1;
  return wait(p);
}

uint64
sys_wait4(void)
{
  int pid;
  uint64 status,result;
  int options;
  if(argint(0, &pid) < 0||argaddr(1, &status) < 0||argint(2, &options) < 0)
    return -1;
  if(pid == -1){
    result = wait(status);
  } else {
    result = wait4(status, pid);
  }
  return result;
}

uint64
sys_sbrk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64 
sys_waitpid(void)
{
  // int pid;
  // uint64 status;
  // int options;

  // // 从用户空间获取 pid、status 和 options
  // if (argint(0, &pid) < 0 || argaddr(1, &status) < 0 || argint(2, &options) < 0) {
  //     return -1;
  // }
  // return wait4(pid,(int*)status,options);
  struct proc *p=myproc();
  long startT=get_uptime();
  yield();
  p->sleep_time+=get_uptime()-startT;
  return 0;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  if(argint(0, &n) < 0)
    return -1;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64 sys_nanosleep(void){
  uint64 p_out_TimeVal;
  struct timeval TimeVal_in;
  struct timeval* p_TimeVal = &TimeVal_in;
  uint ticks0;

  if(argaddr(0, &p_out_TimeVal) < 0)
    return -1;
  
  copyin2((char *)p_TimeVal, p_out_TimeVal, sizeof(TimeVal_in));
 
  int sleep_dida = p_TimeVal->tv_sec * 8;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < sleep_dida){
    // printf("sleeping\n");
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_kill(void)
{
  int pid;

  if(argint(0, &pid) < 0)
    return -1;
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_trace(void)
{
  int mask;
  if(argint(0, &mask) < 0) {
    return -1;
  }
  myproc()->tmask = mask;
  return 0;
}

uint64 
sys_times(void) {
  uint64 mytimes;
  if(argaddr(0,&mytimes)<0){
    return -1;
  }
  struct proc *p=myproc();
  struct tms tm;
  tm.tms_utime=p->usedtime.tms_utime;
  tm.tms_stime=p->usedtime.tms_stime;
  tm.tms_cutime=p->usedtime.tms_cutime;
  tm.tms_cstime=p->usedtime.tms_cstime;
  if(copyout2(mytimes, (char *)&tm, sizeof(tm)) < 0)
      return -1;
  return 0;
}

uint64 
sys_brk(void){
  uint64 newaddr;
  if(argaddr(0,&newaddr)<0)
    return -1;
  uint64 curaddr=myproc()->sz;
  if(newaddr==0)
    return curaddr;
  if(growproc(newaddr-curaddr)<0) 
    return -1;
  return newaddr;
}